import random

import torch
import torch.nn as nn

from advex_uar.attacks.attacks import AttackWrapper

class FrankWolfeAttack(AttackWrapper):
    def __init__(self, nb_its, eps_max, resol, rand_init=True, scale_each=False):
        """
        Parameters:
            nb_its (int):          Number of FW iterations.
            eps_max (float):       The max norm, in pixel space (out of 255). Total L1 norm is (resol * resol * 3 * eps_max)
            resol (int):           Side length of the image.
            rand_init (bool):      Whether to use a random init in the l1 ball
            scale_each (bool):     Whether to scale eps for each image in a batch separately
        """
        super().__init__(resol)
        self.nb_its = nb_its
        self.eps_max = eps_max
        num_pixels = resol * resol * 3
        self.l1_max = eps_max * num_pixels
        self.criterion = nn.CrossEntropyLoss().cuda()
        self.rand_init = rand_init
        self.scale_each = scale_each

        self.nb_backward_steps = self.nb_its

    def _run_one(self, pixel_model, pixel_inp, delta, target, base_eps, avoid_target=True):
        s = pixel_model(pixel_inp + delta)
        for it in range(self.nb_its):
            loss = self.criterion(s, target)
            loss.backward()
            '''
            Because of batching, this grad is scaled down by 1 / batch_size, which does not matter
            for what follows because of normalization.
            '''
            if avoid_target:
                grad = delta.grad.data
            else:
                grad = - delta.grad.data
            
            def where_float(cond, if_true, if_false):
                return cond.float() * if_true + (1-cond.float()) * if_false

            def where_long(cond, if_true, if_false):
                return cond.long() * if_true + (1-cond.long()) * if_false


            # shape of grad: (batch, c, x, y)
            batch_size = grad.size()[0]
            abs_grad = torch.abs(grad).view(batch_size, -1)
            num_pixels = abs_grad.size()[1]
            sign_grad = torch.sign(grad)

            bound = where_float(sign_grad > 0, 255 - pixel_inp, pixel_inp).view(batch_size, -1)
                
            k_min = torch.zeros((batch_size,1), dtype=torch.long, requires_grad=False, device='cuda')
            k_max = torch.ones((batch_size,1), dtype=torch.long, requires_grad=False, device='cuda') * num_pixels
                
            # cum_bnd[k] is meant to track the L1 norm we end up with if we take 
            # the k indices with the largest gradient magnitude and push them to their boundary values (0 or 255)
            values, indices = torch.sort(abs_grad, descending=True)
            bnd = torch.gather(bound, 1, indices)
            # subtract bnd because we don't want the cumsum to include the final element
            cum_bnd = torch.cumsum(bnd, 1) - bnd
                
            # this is hard-coded as floor(log_2(256 * 256 * 3))
            for _ in range(17):
                k_mid = (k_min + k_max) // 2
                l1norms = torch.gather(cum_bnd, 1, k_mid)
                k_min = where_long(l1norms > base_eps, k_min, k_mid)
                k_max = where_long(l1norms > base_eps, k_mid, k_max)
                
            # next want to set the gradient of indices[0:k_min] to their corresponding bound
            magnitudes = torch.zeros((batch_size, num_pixels), requires_grad=False, device='cuda')
            for bi in range(batch_size):
                magnitudes[bi, indices[bi, :k_min[bi,0]]] = bnd[bi, :k_min[bi,0]]
                magnitudes[bi, indices[bi, k_min[bi,0]]] = base_eps[bi] - cum_bnd[bi, k_min[bi,0]]
                
            delta_it = sign_grad * magnitudes.view(pixel_inp.size())
            # These should always be exactly epsilon
            # l1_check = torch.norm(delta_it.view(batch_size, -1), 1.0, dim=1) / num_pixels
            # print('l1_check: %s' % l1_check)
            delta.data = delta.data + (delta_it - delta.data) / (it + 1.0)
            # These will generally be a bit smaller than epsilon
            # l1_check2 = torch.norm(delta.data.view(batch_size, -1), 1.0, dim=1) / num_pixels
            # print('l1_check2: %s' % l1_check2)

            if it != self.nb_its - 1:
                s = pixel_model(pixel_inp + delta)
            delta.grad.data.zero_()
        return delta

    def _init(self, shape, eps):
        if self.rand_init:
            # algorithm from https://arxiv.org/abs/math/0503650, Theorem 1
            exp = torch.empty(shape, dtype=torch.float32, device='cuda')
            exp.exponential_()
            signs = torch.sign(torch.randn(shape, dtype=torch.float32, device='cuda'))
            exp = exp * signs
            exp_y = torch.empty(shape[0], dtype=torch.float32, device='cuda')
            exp_y.exponential_()
            norm = exp_y + torch.norm(exp.view(shape[0], -1), 1.0, dim=1)
            init = exp / norm[:, None, None, None]
            init = eps[:, None, None, None] * init
            init.requires_grad_()
            return init
        else:
            return torch.zeros(shape, requires_grad=True, device='cuda')
    
    def _forward(self, pixel_model, pixel_img, target, scale_eps=False, avoid_target=True):
        if scale_eps:
            if self.scale_each:
                rand = torch.rand(pixel_img.size()[0], device='cuda')
            else:
                rand = random.random() * torch.ones(pixel_img.size()[0], device='cuda')
            base_eps = rand * self.l1_max
        else:
            base_eps = self.l1_max * torch.ones(pixel_img.size()[0], device='cuda')

        pixel_inp = pixel_img.detach()
        pixel_inp.requires_grad = True
        delta = self._init(pixel_inp.size(), base_eps)
        if self.nb_its > 0:
            delta = self._run_one(pixel_model, pixel_inp, delta, target,
                                  base_eps, avoid_target=avoid_target)
        else:
            delta.data = torch.clamp(pixel_inp.data + delta.data, 0., 255.) - pixel_inp.data
        pixel_result = pixel_inp + delta
        return pixel_result




import random

import torch
import torch.nn as nn

from advex_uar.attacks.attacks import AttackWrapper
from advex_uar.attacks.elastic import ElasticDeformation

class ElasticAttack(AttackWrapper):
    def __init__(self, nb_its, eps_max, step_size, resol,
                 rand_init=True, scale_each=False,
                 kernel_size=25, kernel_std=3):
        '''
        Arguments:
            nb_its (int):          Number of iterations
            eps_max (float):       Maximum flow, in L_inf norm, in pixels
            step_size (float):     Maximum step size in L_inf norm, in pixels
            resol (int):           Side length of images, in pixels
            rand_init (bool):      Whether to do a random init
            scale_each (bool):     Whether to scale eps for each image in a batch separately
            kernel_size (int):     Size, in pixels of gaussian kernel
            kernel_std (int):      Standard deviation of kernel
        '''
        super().__init__(resol)
        self.nb_its = nb_its
        self.eps_max = eps_max
        self.step_size = step_size
        self.resol = resol
        self.rand_init = rand_init
        self.scale_each = scale_each

        self.deformer = ElasticDeformation(resol, kernel_size, kernel_std)
        self.criterion = nn.CrossEntropyLoss().cuda()
        self.nb_backward_steps = self.nb_its

    def _init(self, batch_size, eps):
        if self.rand_init:
            # initialized randomly in [-1, 1], then scaled to [-base_eps, base_eps]
            flow = torch.rand((batch_size, 2, self.resol, self.resol),
                              dtype=torch.float32, device='cuda') * 2 - 1
            flow = eps[:, None, None, None] * flow
        else:
            flow = torch.zeros((batch_size, 2, self.resol, self.resol),
                               dtype=torch.float32, device='cuda')
        flow.requires_grad_()
        return flow
        
    def _forward(self, pixel_model, pixel_img, target, scale_eps=False, avoid_target=True):
        pixel_inp = pixel_img.detach()
        pixel_inp.requires_grad = True

        if scale_eps:
            if self.scale_each:
                rand = torch.rand(pixel_img.size()[0], device='cuda')
            else:
                rand = random.random() * torch.ones(pixel_img.size()[0], device='cuda')
            base_eps = rand * self.eps_max
            step_size = rand * self.step_size
        else:
            base_eps = self.eps_max * torch.ones(pixel_img.size()[0], device='cuda')
            step_size = self.step_size * torch.ones(pixel_img.size()[0], device='cuda')

        # Our base_eps and step_size are in pixel scale, but flow is in [-1, 1] scale
        base_eps.mul_(2.0 / self.resol)
        step_size.mul_(2.0 / self.resol)

        flow = self._init(pixel_img.size()[0], base_eps)        
        pixel_inp_adv = self.deformer(pixel_inp, flow)

        if self.nb_its > 0:
            res = pixel_model(pixel_inp_adv)       
            for it in range(self.nb_its):
                loss = self.criterion(res, target)
                loss.backward()

                if avoid_target:
                    grad = flow.grad.data
                else:
                    grad = -flow.grad.data
                
                # step_size has already been converted to [-1, 1] scale
                flow.data = flow.data + step_size[:, None, None, None] * grad.sign()
                flow.data = torch.max(torch.min(flow.data, base_eps[:, None, None, None]), -base_eps[:, None, None, None])            
                pixel_inp_adv = self.deformer(pixel_inp, flow)
                if it != self.nb_its - 1:
                    res = pixel_model(pixel_inp_adv)
                    flow.grad.data.zero_()
        return pixel_inp_adv




import datetime
import functools
import math
import os
import uuid

from torchvision import models

from advex_uar.attacks import PGDAttack, ElasticAttack, FrankWolfeAttack,\
    JPEGAttack, GaborAttack, FogAttack, SnowAttack
from advex_uar.common.models import cifar10_resnet
from advex_uar.logging.logger import Logger

def init_logger(use_wandb, job_type, flags):
    if use_wandb:
        log_dir = None
        run_id = None
    else:
        dir_path = os.getcwd()
        time_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        run_id = str(uuid.uuid4())[:8]
        dir_str = '{}-{}-{}'.format(job_type, time_str, run_id)
        log_dir = os.path.join(dir_path, job_type, dir_str)
        os.makedirs(log_dir, exist_ok=True)
    logger = Logger(use_wandb, job_type, run_id, log_dir=log_dir, flags=flags)
    return logger
        
def get_imagenet_model(resnet_size, nb_classes):
    size_to_model = {
        18: models.resnet18,
        34: models.resnet34,
        50: models.resnet50,
        101: models.resnet101,
        152: models.resnet152
    }
    return size_to_model[resnet_size](num_classes=nb_classes)

def get_cifar10_model(resnet_size):
    size_to_model = {
        20: cifar10_resnet.resnet20,
        32: cifar10_resnet.resnet32,
        44: cifar10_resnet.resnet44,
        56: cifar10_resnet.resnet56,
        110: cifar10_resnet.resnet110,
    }
    model = size_to_model[resnet_size]()
    return model

def get_model(dataset, resnet_size, nb_classes):
    if dataset in ['imagenet', 'imagenet-c']:
        return get_imagenet_model(resnet_size, nb_classes)
    elif dataset in ['cifar-10', 'cifar-10-c']:
        return get_cifar10_model(resnet_size)

def _get_attack(dataset, attack, eps, n_iters, step_size, scale_each):
    if dataset in ['imagenet', 'imagenet-c']:
        resol = 224
        elastic_kernel = 25
        elastic_std = 3
    elif dataset in ['cifar-10', 'cifar-10-c']:
        resol = 32
        elastic_kernel = 5
        elastic_std = 3.0/224.0 * 32

    if attack == 'pgd_linf':
        return functools.partial(PGDAttack, n_iters, eps,
                                 step_size, resol, norm='linf',
                                 scale_each=scale_each)
    elif attack == 'pgd_l2':
        return functools.partial(PGDAttack, n_iters, eps,
                                 step_size, resol, norm='l2',
                                 scale_each=scale_each)
    elif attack == 'fw_l1':
        return functools.partial(FrankWolfeAttack, n_iters, eps,
                                 resol, scale_each=scale_each)
    elif attack == 'jpeg_linf':
        return functools.partial(JPEGAttack, n_iters, eps,
                                 step_size, resol, scale_each=scale_each,
                                 opt='linf')
    elif attack == 'jpeg_l2':
        return functools.partial(JPEGAttack, n_iters, eps,
                                 step_size, resol, scale_each=scale_each,
                                 opt='l2')
    elif attack == 'jpeg_l1':
        return functools.partial(JPEGAttack, n_iters, eps,
                                 step_size, resol, scale_each=scale_each,
                                 opt='l1')
    elif attack == 'elastic':
        return functools.partial(ElasticAttack, n_iters, eps,
                                 step_size, resol, scale_each=scale_each,
                                 kernel_size=elastic_kernel,
                                 kernel_std=elastic_std)
    elif attack == 'fog':
        return functools.partial(FogAttack, n_iters, eps,
                                 step_size, resol, scale_each=scale_each)
    elif attack == 'gabor':
        return functools.partial(GaborAttack, n_iters, eps,
                                 step_size, resol, scale_each=scale_each)
    elif attack == 'snow':
        return functools.partial(SnowAttack, n_iters, eps,
                                 step_size, resol, scale_each=scale_each)
    else:
        raise NotImplementedError    
    
def get_attack(*args, **kwargs):
    return _get_attack(*args, **kwargs)

def get_step_size(epsilon, n_iters, use_max=False):
    if use_max:
        return epsilon
    else:
        return epsilon / math.sqrt(n_iters)



class FlagHolder(object):
    def __init__(self):
        self._dict = None

    def initialize(self, **kwargs):
        self._dict = {}
        for k, v in kwargs.items():
            self._dict[k] = v

    def summary(self):
        print('===== Flag summary =====')
        for k, v in self._dict.items():
            print('{k}: {v}'.format(k=k, v=v))
        print('=== End flag summary ===')

    def __getattr__(self, key):
        return self._dict[key]



import importlib
import os
import subprocess

from PIL import Image
import numpy as np

import torch
import torch.nn.functional as F
import torch.utils.data
from torchvision import datasets, transforms, models

from advex_uar.common.loader import StridedImageFolder
from advex_uar.eval.cifar10c import CIFAR10C
from advex_uar.train.trainer import Metric, accuracy, correct

IMAGENET_MEAN = [0.485, 0.456, 0.406]
IMAGENET_STD = [0.229, 0.224, 0.225]

def norm_to_pil_image(img):
    img_new = torch.Tensor(img)
    for t, m, s in zip(img_new, IMAGENET_MEAN, IMAGENET_STD):
        t.mul_(s).add_(m)
    img_new.mul_(255)
    np_img = np.rollaxis(np.uint8(img_new.numpy()), 0, 3)
    return Image.fromarray(np_img, mode='RGB')

class Accumulator(object):
    def __init__(self, name):
        self.name = name
        self.vals = []

    def update(self, val):
        self.vals.append(val)

    @property
    def avg(self):
        total_sum = sum([torch.sum(v) for v in self.vals])
        total_size = sum([v.size()[0] for v in self.vals])
        return total_sum / total_size

class BaseEvaluator():
    def __init__(self, **kwargs):
        default_attr = dict(
            # eval options
            model=None, batch_size=32, stride=10,
            dataset_path=None, # val dir for imagenet, base dir for CIFAR-10-C
            nb_classes=None,
            # attack options
            attack=None,
            # Communication options
            fp16_allreduce=False,
            # Logging options
            logger=None)
        default_attr.update(kwargs)
        for k in default_attr:
            setattr(self, k, default_attr[k])
        if self.dataset not in ['imagenet', 'imagenet-c', 'cifar-10', 'cifar-10-c']:
            raise NotImplementedError
        self.cuda = True
        if self.cuda:
            self.model.cuda()
        self.attack = self.attack()
        self._init_loaders()

    def _init_loaders(self):
        raise NotImplementedError
        
    def evaluate(self):
        self.model.eval()

        std_loss = Accumulator('std_loss')
        adv_loss = Accumulator('adv_loss')
        std_corr = Accumulator('std_corr')
        adv_corr = Accumulator('adv_corr')
        std_logits = Accumulator('std_logits')
        adv_logits = Accumulator('adv_logits')

        seen_classes = []
        adv_images = Accumulator('adv_images')
        first_batch_images = Accumulator('first_batch_images')

        for batch_idx, (data, target) in enumerate(self.val_loader):
            if self.cuda:
                data, target = data.cuda(non_blocking=True), target.cuda(non_blocking=True)
            with torch.no_grad():
                output = self.model(data)
                std_logits.update(output.cpu())
                loss = F.cross_entropy(output, target, reduction='none').cpu()
                std_loss.update(loss)
                corr = correct(output, target)
                corr = corr.view(corr.size()[0]).cpu()
                std_corr.update(corr)

            rand_target = torch.randint(
                0, self.nb_classes - 1, target.size(),
                dtype=target.dtype, device='cuda')
            rand_target = torch.remainder(target + rand_target + 1, self.nb_classes)
            data_adv = self.attack(self.model, data, rand_target,
                                   avoid_target=False, scale_eps=False)

            for idx in range(target.size()[0]):
                if target[idx].cpu() not in seen_classes:
                    seen_classes.append(target[idx].cpu())
                    orig_image = norm_to_pil_image(data[idx].detach().cpu())
                    adv_image = norm_to_pil_image(data_adv[idx].detach().cpu())
                    adv_images.update((orig_image, adv_image, target[idx].cpu()))

            if batch_idx == 0:
                for idx in range(target.size()[0]):
                    orig_image = norm_to_pil_image(data[idx].detach().cpu())
                    adv_image = norm_to_pil_image(data_adv[idx].detach().cpu())
                    first_batch_images.update((orig_image, adv_image))
                
            with torch.no_grad():
                output_adv = self.model(data_adv)
                adv_logits.update(output_adv.cpu())
                loss = F.cross_entropy(output_adv, target, reduction='none').cpu()
                adv_loss.update(loss)
                corr = correct(output_adv, target)
                corr = corr.view(corr.size()[0]).cpu()
                adv_corr.update(corr)

            run_output = {'std_loss':std_loss.avg,
                          'std_acc':std_corr.avg,
                          'adv_loss':adv_loss.avg,
                          'adv_acc':adv_corr.avg}
            print('Batch', batch_idx)
            print(run_output)
            if batch_idx % 20 == 0:
                self.logger.log(run_output, batch_idx)

        summary_dict = {'std_acc':std_corr.avg.item(),
                        'adv_acc':adv_corr.avg.item()}
        self.logger.log_summary(summary_dict)
        for orig_img, adv_img, target in adv_images.vals:
            self.logger.log_image(orig_img, 'orig_{}.png'.format(target))
            self.logger.log_image(adv_img, 'adv_{}.png'.format(target))
        for idx, imgs in enumerate(first_batch_images.vals):
            orig_img, adv_img = imgs
            self.logger.log_image(orig_img, 'init_orig_{}.png'.format(idx))
            self.logger.log_image(adv_img, 'init_adv_{}.png'.format(idx))

        self.logger.end()
        print(std_loss.avg, std_corr.avg, adv_loss.avg, adv_corr.avg)

class CIFAR10Evaluator(BaseEvaluator):
    def _init_loaders(self):
        normalize = transforms.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD)
        self.val_dataset = datasets.CIFAR10(
                root='./', download=True, train=False,
                transform=transforms.Compose([
                        transforms.ToTensor(),
                        normalize,]))
        self.val_loader = torch.utils.data.DataLoader(
                self.val_dataset, batch_size=self.batch_size,
                shuffle=False, num_workers=8, pin_memory=True)

class ImagenetEvaluator(BaseEvaluator):
    def _init_loaders(self):
        valdir = self.dataset_path
        normalize = transforms.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD)
        self.val_dataset = StridedImageFolder(
                valdir,
                transforms.Compose([
                        transforms.Resize(256),
                        transforms.CenterCrop(224),
                        transforms.ToTensor(),
                        normalize,]),
                stride=self.stride)
        self.val_sampler = torch.utils.data.SequentialSampler(self.val_dataset)
        self.val_loader = torch.utils.data.DataLoader(
                self.val_dataset, batch_size=self.batch_size,
                sampler=self.val_sampler, num_workers=1, pin_memory=True,
                shuffle=False)

class ImagenetCEvaluator(BaseEvaluator):
    def __init__(self, corruption_type=None, corruption_name=None, corruption_level=None, **kwargs):
        self.corruption_type = corruption_type
        self.corruption_name = corruption_name
        self.corruption_level = corruption_level
        super().__init__(**kwargs)
    
    def _init_loaders(self):
        valdir = os.path.join(self.dataset_path, 'imagenet-c',
                              self.corruption_type, self.corruption_name, self.corruption_level)
        normalize = transforms.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD)
        self.val_dataset = StridedImageFolder(
                valdir,
                transforms.Compose([
                        transforms.Resize(256),
                        transforms.CenterCrop(224),
                        transforms.ToTensor(),
                        normalize,]),
                stride=self.stride)
        self.val_sampler = torch.utils.data.SequentialSampler(self.val_dataset)
        self.val_loader = torch.utils.data.DataLoader(
                self.val_dataset, batch_size=self.batch_size,
                sampler=self.val_sampler, num_workers=1, pin_memory=True,
                shuffle=False)

class CIFAR10CEvaluator(BaseEvaluator):
    def __init__(self, corruption_type=None, corruption_name=None, corruption_level=None, **kwargs):
        self.corruption_type = corruption_type
        self.corruption_name = corruption_name
        self.corruption_level = corruption_level
        super().__init__(**kwargs)
    
    def _init_loaders(self):
        valdir = os.path.join(self.dataset_path, 'CIFAR-10-C')
        transform = transforms.Compose(
                [transforms.ToTensor(),
                 transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                      std=[0.229, 0.224, 0.225])])
        self.val_dataset = CIFAR10C(valdir, transform=transform,
                                    corruption_name=self.corruption_name,
                                    corruption_level=self.corruption_level)
        self.val_sampler = torch.utils.data.SequentialSampler(self.val_dataset)
        self.val_loader = torch.utils.data.DataLoader(
                self.val_dataset, batch_size=self.batch_size,
                sampler=self.val_sampler, num_workers=1, pin_memory=True,
                shuffle=False)



import random

import numpy as np
import torch
import torch.nn as nn

from advex_uar.attacks.attacks import AttackWrapper
from advex_uar.attacks.fog import fog_creator

class FogAttack(AttackWrapper):
    def __init__(self, nb_its, eps_max, step_size, resol, rand_init=True, scale_each=False,
                 wibble_decay=2.0):
        """
        Parameters:
            nb_its (int):          Number of GD iterations.
            eps_max (float):       The max norm, in pixel space.
            step_size (float):     The max step size, in pixel space.
            resol (int):           Side length of the image.
            rand_init (bool):      Whether to init randomly in the norm ball
            scale_each (bool):     Whether to scale eps for each image in a batch separately
            wibble_decay (float):  Fog-specific parameter
        """
        super().__init__(resol)
        self.nb_its = nb_its
        self.eps_max = eps_max
        self.step_size = step_size
        self.resol = resol
        self.rand_init = rand_init
        self.scale_each = scale_each
        self.wibble_decay = wibble_decay

        self.criterion = nn.CrossEntropyLoss().cuda()
        self.nb_backward_steps = self.nb_its    

    def _init(self, batch_size, map_size):
        fog_vars = []
        for i in range(int(np.log2(map_size))):
            for j in range(3):
                var = torch.rand((batch_size, 2**i, 2**i), device="cuda")
                var.requires_grad_()
                fog_vars.append(var)
        return fog_vars
        
    def _forward(self, pixel_model, pixel_img, target, avoid_target=True, scale_eps=False):
        pixel_inp = pixel_img.detach()
        batch_size = pixel_img.size(0)
        x_max, _ = torch.max(pixel_img.view(pixel_img.size(0), 3, -1), -1)
        x_max = x_max.view(-1, 3, 1, 1)
        map_size = 2 ** (int(np.log2(self.resol)) + 1)
        
        if scale_eps:
            if self.scale_each:
                rand = torch.rand(pixel_img.size()[0], device='cuda')
            else:
                rand = random.random() * torch.ones(pixel_img.size()[0], device='cuda')
            base_eps = rand.mul(self.eps_max)
            step_size = self.step_size * torch.ones(pixel_img.size()[0], device='cuda')
        else:
            base_eps = self.eps_max * torch.ones(pixel_img.size()[0], device='cuda')
            step_size = self.step_size * torch.ones(pixel_img.size()[0], device='cuda')
        
        fog_vars = self._init(batch_size, map_size)
        fog = fog_creator(fog_vars, batch_size, mapsize=map_size,
                          wibbledecay=self.wibble_decay)[:,:,16:-16,16:-16]
        s = pixel_model(torch.clamp((pixel_inp + base_eps[:, None, None, None] * fog) /
                                    (x_max + base_eps[:, None, None, None]) * 255., 0., 255.))
        for it in range(self.nb_its):
            loss = self.criterion(s, target)
            loss.backward()
            '''
            Because of batching, this grad is scaled down by 1 / batch_size, which does not matter
            for what follows because of normalization.
            '''
            if avoid_target:
                # to avoid the target, we increase the loss
                grads = [f.grad.data for f in fog_vars]
            else:
                # to hit the target, we reduce the loss
                grads = [-f.grad.data for f in fog_vars]

            grad_signs = [grad.sign() for grad in grads]
            for f, g in zip(fog_vars, grad_signs):
                f.data = f.data + step_size[:, None, None] * g
                f.detach()
                f.data = f.data.clamp(0, 1)
                f.requires_grad_()

            if it != self.nb_its - 1:
                fog = fog_creator(fog_vars, batch_size, mapsize=map_size,
                                  wibbledecay=self.wibble_decay)[:,:,16:-16,16:-16]
                s = pixel_model(torch.clamp((pixel_inp + base_eps[:, None, None, None] * fog) /
                                            (x_max + base_eps[:, None, None, None]) * 255., 0., 255.))
                for f in fog_vars:
                    f.grad.data.zero_()
        fog = fog_creator(fog_vars, batch_size, mapsize=map_size,
                          wibbledecay=self.wibble_decay)[:,:,16:-16,16:-16]
        pixel_result = torch.clamp((pixel_inp + base_eps[:, None, None, None] * fog) /
                                   (x_max + base_eps[:, None, None, None]) * 255., 0., 255.)
        return pixel_result



import math
import numbers

import torch
import numpy as np
from torch import nn
from torch.nn import functional as F

# Taken from: https://discuss.pytorch.org/t/is-there-anyway-to-do-gaussian-filtering-for-an-image-2d-3d-in-pytorch/12351/10
class GaussianSmoothing(nn.Module):
    """
    Apply gaussian smoothing on a
    1d, 2d or 3d tensor. Filtering is performed seperately for each channel
    in the input using a depthwise convolution.
    Arguments:
        channels (int, sequence): Number of channels of the input tensors. Output will
            have this number of channels as well.
        kernel_size (int, sequence): Size of the gaussian kernel.
        sigma (float, sequence): Standard deviation of the gaussian kernel.
        dim (int, optional): The number of dimensions of the data.
            Default value is 2 (spatial).
    """
    def __init__(self, channels, kernel_size, sigma, dim=2):
        super(GaussianSmoothing, self).__init__()
        if isinstance(kernel_size, numbers.Number):
            kernel_size = [kernel_size] * dim
        if isinstance(sigma, numbers.Number):
            sigma = [sigma] * dim

        # The gaussian kernel is the product of the
        # gaussian function of each dimension.
        kernel = 1
        meshgrids = torch.meshgrid(
            [
                torch.arange(size, dtype=torch.float32)
                for size in kernel_size
            ]
        )
        for size, std, mgrid in zip(kernel_size, sigma, meshgrids):
            mean = (size - 1) / 2
            kernel *= 1 / (std * math.sqrt(2 * math.pi)) * \
                      torch.exp(-((mgrid - mean) / std) ** 2 / 2)

        # Make sure sum of values in gaussian kernel equals 1.
        kernel = kernel / torch.sum(kernel)

        # Reshape to depthwise convolutional weight
        kernel = kernel.view(1, 1, *kernel.size())
        kernel = kernel.repeat(channels, *[1] * (kernel.dim() - 1))

        self.register_buffer('weight', kernel)
        self.groups = channels

        if dim == 1:
            self.conv = F.conv1d
        elif dim == 2:
            self.conv = F.conv2d
        elif dim == 3:
            self.conv = F.conv3d
        else:
            raise RuntimeError(
                'Only 1, 2 and 3 dimensions are supported. Received {}.'.format(dim)
            )

    def forward(self, inp):
        """
        Apply gaussian filter to input.
        Arguments:
            input (torch.Tensor): Input to apply gaussian filter on.
        Returns:
            filtered (torch.Tensor): Filtered output.
        """
        return self.conv(inp, weight=self.weight, groups=self.groups)

class ElasticDeformation(nn.Module):
    def __init__(self, im_size, filter_size, std):
        super().__init__()
        self.im_size = im_size
        self.filter_size = filter_size
        self.std = std
        self.kernel = GaussianSmoothing(2, self.filter_size, self.std).cuda()

        self._get_base_flow()

    def _get_base_flow(self):
        xflow, yflow = np.meshgrid(
                np.linspace(-1, 1, self.im_size, dtype='float32'),
                np.linspace(-1, 1, self.im_size, dtype='float32'))
        flow = np.stack((xflow, yflow), axis=-1)
        flow = np.expand_dims(flow, axis=0)
        self.base_flow = nn.Parameter(torch.from_numpy(flow)).cuda().detach()

    def warp(self, im, flow):
        return F.grid_sample(im, flow, mode='bilinear')

    def forward(self, im, params):
        flow = F.pad(params, ((self.filter_size - 1) // 2, ) * 4 , mode='reflect')
        local_flow = self.kernel(flow).transpose(1, 3).transpose(1, 2)
        return self.warp(im, local_flow + self.base_flow)



from io import BytesIO
import os

import numpy as np
from PIL import Image
import torch

class CIFAR10C(torch.utils.data.Dataset):
    """`CIFAR10-C`_ Dataset.
    Args:
        data_dir (string): Directory containing .npy files
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        corruption_name (str): Name of the corruption
        corruption_level (int): Level of the corruption (1, 2, 3, 4, 5)
        apply_compress (bool): Whether to apply JPEG compression
   """
    def __init__(self, data_dir, transform=None, corruption_name=None, corruption_level=None,
                 apply_compress=False):
        super(CIFAR10C, self).__init__()
        self.transform = transform
        self.corruption_name = corruption_name
        self.corruption_level = int(corruption_level)
        self.apply_compress = apply_compress
        
        self.data = np.load(os.path.join(data_dir, corruption_name + '.npy'))
        self.targets = np.load(os.path.join(data_dir, 'labels.npy')).astype(int)

        self.data = self.data[10000 * (self.corruption_level - 1): 10000 * self.corruption_level]
        self.targets = self.targets[10000 * (self.corruption_level - 1): 10000 * self.corruption_level]

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)
        if self.apply_compress:
            f = BytesIO()
            img.save(f, format='JPEG', quality=85, optimize=True)
            f.seek(0)
            img = Image.open(f)
        
        if self.transform is not None:
            img = self.transform(img)

        return img, target

    def __len__(self):
        return len(self.data)



import math
import numbers
import numpy as np
import torch
import torch.nn.functional as F
import torch.nn as nn

# modification of code from pytorch forums
class GaussianSmoothing(nn.Module):
    """
    Apply gaussian smoothing on a tensor. Filtering is performed seperately for each channel
    in the input using a depthwise convolution.
    Arguments:
        channels (int, sequence): Number of channels of the input tensors. Output will
            have this number of channels as well.
        kernel_size (int, sequence): Size of the gaussian kernel.
        sigma (float, sequence): Standard deviation of the gaussian kernel.
    """
    def __init__(self, kernel_size, sigma, channels=1):
        super(GaussianSmoothing, self).__init__()
        if isinstance(kernel_size, numbers.Number):
            kernel_size = [kernel_size] * 2
        if isinstance(sigma, numbers.Number):
            sigma = [sigma] * 2

        # The gaussian kernel is the product of the
        # gaussian function of each dimension.
        kernel = 1
        meshgrids = torch.meshgrid(
            [
                torch.arange(size, dtype=torch.float32)
                for size in kernel_size
            ]
        )
        for size, std, mgrid in zip(kernel_size, sigma, meshgrids):
            mean = (size - 1) / 2
            kernel *= 1 / (std * math.sqrt(2 * math.pi)) * \
                      torch.exp(-((mgrid - mean) / std) ** 2 / 2)

        # Make sure sum of values in gaussian kernel equals 1.
        kernel = kernel / torch.sum(kernel)

        # Reshape to depthwise convolutional weight
        kernel = kernel.view(1, 1, *kernel.size())
        kernel = kernel.repeat(channels, *[1] * (kernel.dim() - 1)).cuda()

        self.register_buffer('weight', kernel)
        self.groups = channels

        self.conv = F.conv2d

    def forward(self, input, padding=0):
        """
        Apply gaussian filter to input.
        Arguments:
            input (torch.Tensor): Input to apply gaussian filter on.
        Returns:
            filtered (torch.Tensor): Filtered output.
        """
        return self.conv(input, weight=self.weight, groups=self.groups, padding=padding)


# from stackoverflow
def trapez(y,y0,w):
    return np.clip(np.minimum(y+1+w/2-y0, -y+1+w/2+y0),0,1)


def weighted_line(r0, c0, r1, c1, w, rmin=0, rmax=np.inf):
    # The algorithm below works fine if c1 >= c0 and c1-c0 >= abs(r1-r0).
    # If either of these cases are violated, do some switches.
    if abs(c1-c0) < abs(r1-r0):
        # Switch x and y, and switch again when returning.
        xx, yy, val = weighted_line(c0, r0, c1, r1, w, rmin=rmin, rmax=rmax)
        return (yy, xx, val)

    # At this point we know that the distance in columns (x) is greater
    # than that in rows (y). Possibly one more switch if c0 > c1.
    if c0 > c1:
        return weighted_line(r1, c1, r0, c0, w, rmin=rmin, rmax=rmax)

    # The following is now always < 1 in abs
    slope = (r1-r0) / (c1-c0)

    # Adjust weight by the slope
    w *= np.sqrt(1+np.abs(slope)) / 2

    # We write y as a function of x, because the slope is always <= 1
    # (in absolute value)
    x = np.arange(c0, c1+1, dtype=float)
    y = x * slope + (c1*r0-c0*r1) / (c1-c0)

    # Now instead of 2 values for y, we have 2*np.ceil(w/2).
    # All values are 1 except the upmost and bottommost.
    thickness = np.ceil(w/2)
    yy = (np.floor(y).reshape(-1,1) + np.arange(-thickness-1,thickness+2).reshape(1,-1))
    xx = np.repeat(x, yy.shape[1])
    vals = trapez(yy, y.reshape(-1,1), w).flatten()

    yy = yy.flatten()

    # Exclude useless parts and those outside of the interval
    # to avoid parts outside of the picture
    mask = np.logical_and.reduce((yy >= rmin, yy < rmax, vals > 0))

    return (yy[mask].astype(int), xx[mask].astype(int), vals[mask])


def make_kernels(snow_length_bound=13, blur=True):
    kernels = []

    flip = np.random.uniform() < 0.5

    for i in range(7):
        k_size = snow_length_bound
        mid = k_size//2
        k_npy = np.zeros((k_size, k_size))
        rr, cc, val = weighted_line(
            mid, mid, np.random.randint(mid+2,k_size), np.random.randint(mid+2,k_size),
            np.random.choice([1,3,5], p=[0.6, 0.3, 0.1]), mid, k_size)

        k_npy[rr, cc] = val
        k_npy[:mid+1, :mid+1] = k_npy[::-1,::-1][:mid+1,:mid+1]

        if flip:
            k_npy = k_npy[:, ::-1]

        kernel = torch.FloatTensor(k_npy.copy()).view(1,1,k_size,k_size).cuda()

        if blur:
            blurriness = np.random.uniform(0.41, 0.6)
            gaussian_blur = GaussianSmoothing(int(np.ceil(5 * blurriness)), blurriness)
            kernel = gaussian_blur(kernel, padding=1)
        kernels.append(kernel)

    return kernels


def snow_creator(intensities, k, resol):
    flake_grids = []
    k = torch.cat(k, 1)

    intensities_pow = torch.pow(intensities, 4)
    flake_grids = torch.zeros((intensities.size(0), k.size(1), resol, resol)).cuda()

    for i in range(4):
        flake_grids[:, i, ::4,i::4] = intensities_pow[:,i]
    for i in range(3):
        flake_grids[:, i+4, i+1::4,::4] = intensities_pow[:,4+i]

    snow = F.conv2d(flake_grids, k, padding=k.size(-1)//2)

    return snow



import importlib
import os
import time

import torch
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data.distributed
import torchvision
from torchvision import datasets, transforms, models
import horovod.torch as hvd

from advex_uar.common.loader import StridedImageFolder

# Horovod: average metrics from distributed training.
class Metric(object):
    def __init__(self, name):
        self.name = name
        self.sum = torch.tensor(0.)
        self.n = torch.tensor(0.)

    def update(self, val):
        self.sum += hvd.allreduce(val.detach().cpu(), name=self.name)
        self.n += 1

    @property
    def avg(self):
        return self.sum / self.n

def accuracy(output, target):
    # get the index of the max log-probability
    pred = output.max(1, keepdim=True)[1]
    return pred.eq(target.view_as(pred)).cpu().float().mean()

def correct(output, target):
    # get the index of the max log-probability
    pred = output.max(1, keepdim=True)[1]
    return pred.eq(target.view_as(pred)).cpu().float()

class BaseTrainer():
    # Notes:
    # The attack needs to be initialized after the cuda device is set
    def __init__(self, **kwargs):
        default_attr = dict(
            # Training options
            batch_size=32, base_lr=0.0125, momentum=0.9, wd=1e-4, epochs=90, warmup_epochs=5,
            stride=10, label_smoothing=-1.0, rand_target=False,
            # validation options
            run_val=True,
            # Model/checkpoint options
            model=None, checkpoint_dir=None, dataset_path='/mnt/imagenet-test/',
            # Attack options
            attack=None, attack_backward_steps=0, attack_loss='avg', scale_eps=False, rand_init=True,
            # Communication options
            fp16_allreduce=False,
            # Logging options
            logger=None)
        default_attr.update(kwargs)
        for k in default_attr:
            setattr(self, k, default_attr[k])
        assert self.attack_loss in ['avg', 'adv_only', 'logsumexp', 'max']

        # Validate args
        assert self.model != None

        # Set up checkpointing
        if self.checkpoint_dir is not None:
            os.makedirs(self.checkpoint_dir, exist_ok=True)

        self.cuda = True
        self.batches_per_allreduce = 1
        self.verbose = 1 if hvd.rank() == 0 else 0
        self.compression = hvd.Compression.fp16 if self.fp16_allreduce else hvd.Compression.none

        if self.verbose:
            print(self.model)

        torch.cuda.set_device(hvd.local_rank())

        if self.cuda:
            self.model.cuda()

        if self.attack:
            self.attack = self.attack()
            self.attack_backward_steps = self.attack.nb_backward_steps

        self.normalize = transforms.Normalize(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

        self._init_loaders()
        self._init_optimizer()
        self._start_sync()

    def _init_loaders(self):
        raise NotImplementedError

    def _init_optimizer(self):
        optimizer = optim.SGD(
                self.model.parameters(),
                lr=(self.base_lr * self.batches_per_allreduce * hvd.size()),
                momentum=self.momentum, weight_decay=self.wd)

        steps_per_batch = self.attack_backward_steps + 1 + 100000000
        if self.attack_loss == 'avg':
            steps_per_batch += 1
        self.optimizer = hvd.DistributedOptimizer(
            optimizer, named_parameters=self.model.named_parameters(),
            compression=self.compression,
            backward_passes_per_step=steps_per_batch * self.batches_per_allreduce)

    def _start_sync(self):
        hvd.broadcast_parameters(self.model.state_dict(), root_rank=0)
        hvd.broadcast_optimizer_state(self.optimizer, root_rank=0)

    def _adjust_learning_rate(self, epoch, batch_idx):
        if epoch < self.warmup_epochs:
            epoch += float(batch_idx + 1) / len(self.train_loader)
            lr_adj = 1. / hvd.size() * (epoch * (hvd.size() - 1) / self.warmup_epochs + 1)
        elif epoch < 30:
            lr_adj = 1.
        elif epoch < 60:
            lr_adj = 1e-1
        elif epoch < 80:
            lr_adj = 1e-2
        else:
            lr_adj = 1e-3
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = self.base_lr * hvd.size() * self.batches_per_allreduce * lr_adj

    def _checkpoint(self, epoch):
        if self.checkpoint_dir and hvd.local_rank() == 0:
            out_fname = '{:02d}.pth'.format(epoch)
            out_fname = os.path.join(self.checkpoint_dir, out_fname)
            state = {'model': self.model.state_dict(),
                     'optimizer': self.optimizer.state_dict()}
            torch.save(state, out_fname)

    def _compute_loss(self, output, target):
        if self.label_smoothing > 0:
            n_class = len(self.val_dataset.classes)
            one_hot = torch.zeros_like(output).scatter(1, target.view(-1, 1), 1)
            one_hot = one_hot.clamp(self.label_smoothing / (n_class - 1), 1 - self.label_smoothing)
            log_prob = F.log_softmax(output, dim=1)
            loss = -(one_hot * log_prob).sum(dim=1)
            return loss.mean()
        else:
            return F.cross_entropy(output, target)
        
    def _train_epoch(self, epoch):
        self.model.train()

        train_std_loss = Metric('train_std_loss')
        train_std_acc = Metric('train_std_acc')
        train_adv_loss = Metric('train_adv_loss')
        train_adv_acc = Metric('train_adv_acc')

        if self.attack:
            self.attack.set_epoch(epoch)
            
        for batch_idx, (data, target) in enumerate(self.train_loader):
            if self.cuda:
                data, target = data.cuda(non_blocking=True), target.cuda(non_blocking=True)
            self._adjust_learning_rate(epoch, batch_idx)
            loss = torch.zeros([], dtype=torch.float32, device='cuda')
            if (not self.attack) or self.attack_loss == 'avg':
                output = self.model(data)
                loss += self._compute_loss(output, target)
                train_std_loss.update(loss)
                train_std_acc.update(accuracy(output, target))
            else:
                with torch.no_grad():
                    self.model.eval()
                    output = self.model(data)
                    train_std_loss_val = self._compute_loss(output, target)            
                    train_std_loss.update(train_std_loss_val)
                    train_std_acc.update(accuracy(output, target))
                    self.model.train()

            if self.attack:
                if self.rand_target:
                    attack_target = torch.randint(
                        0, len(self.val_dataset.classes) - 1, target.size(),
                        dtype=target.dtype, device='cuda')
                    attack_target = torch.remainder(target + attack_target + 1, len(self.val_dataset.classes))
                adv_loss = torch.zeros([], dtype=torch.float32, device='cuda')
                if self.rand_target:
                    data_adv = self.attack(self.model, data, attack_target,
                                           avoid_target=False, scale_eps=self.scale_eps)
                else:
                    data_adv = self.attack(self.model, data, target,
                                           avoid_target=True, scale_eps=self.scale_eps)
                output_adv = self.model(data_adv)
                adv_loss = self._compute_loss(output_adv, target)
                
                train_adv_loss.update(adv_loss)
                train_adv_acc.update(accuracy(output_adv, target))
                loss += adv_loss
                if self.attack_loss == 'avg':
                    loss /= 2.

            self.optimizer.synchronize()
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

        if hvd.rank() == 0:
            log_dict = {'train_std_loss':train_std_loss.avg.item(),
                        'train_std_acc':train_std_acc.avg.item(),
                        'train_adv_loss':train_adv_loss.avg.item(),
                        'train_adv_acc':train_adv_acc.avg.item()}
            print(log_dict)
            self.logger.log(log_dict, epoch)

    def _val_epoch(self, epoch):
        self.model.eval()

        val_std_loss = Metric('val_std_loss')
        val_std_acc = Metric('val_std_acc')

        val_adv_acc = Metric('val_adv_acc')
        val_adv_loss = Metric('val_adv_loss')
        val_max_adv_acc = Metric('val_max_adv_acc')
        val_max_adv_loss = Metric('val_max_adv_loss')

        for batch_idx, (data, target) in enumerate(self.val_loader):
            if self.cuda:
                data, target = data.cuda(non_blocking=True), target.cuda(non_blocking=True)
            with torch.no_grad():
                output = self.model(data)
                val_std_loss.update(F.cross_entropy(output, target))
                val_std_acc.update(accuracy(output, target))
            if self.attack:
                rand_target = torch.randint(
                    0, len(self.val_dataset.classes) - 1, target.size(),
                    dtype=target.dtype, device='cuda')
                rand_target = torch.remainder(target + rand_target + 1, len(self.val_dataset.classes))
                data_adv = self.attack(self.model, data, rand_target,
                                       avoid_target=False, scale_eps=self.scale_eps)
                data_max_adv = self.attack(self.model, data, rand_target, avoid_target=False, scale_eps=False)
                with torch.no_grad():
                    output_adv = self.model(data_adv)
                    val_adv_loss.update(F.cross_entropy(output_adv, target))
                    val_adv_acc.update(accuracy(output_adv, target))
                    
                    output_max_adv = self.model(data_max_adv)
                    val_max_adv_loss.update(F.cross_entropy(output_max_adv, target))
                    val_max_adv_acc.update(accuracy(output_max_adv, target))
            self.model.eval()

        if hvd.rank() == 0:
            log_dict = {'val_std_loss':val_std_loss.avg.item(),
                        'val_std_acc':val_std_acc.avg.item(),
                        'val_adv_loss':val_adv_loss.avg.item(),
                        'val_adv_acc':val_adv_acc.avg.item(),
                        'val_adv_loss':val_max_adv_loss.avg.item(),
                        'val_max_adv_acc':val_max_adv_acc.avg.item()}
            self.logger.log(log_dict, epoch)

        if self.verbose:
            print(log_dict)

        self.optimizer.synchronize()
        self.optimizer.zero_grad()

    def train(self):
        if hvd.rank() == 0:
            print('Beginning training with {} epochs'.format(self.epochs))
        for epoch in range(self.epochs):
            begin = time.time()
            self._train_epoch(epoch)
            if self.run_val:
                self._val_epoch(epoch)
            self._checkpoint(epoch)
            end = time.time()
            if self.verbose:
                print('Epoch {} took {:.2f} seconds'.format(epoch, end - begin))
        if hvd.rank() == 0:
            self.logger.log_ckpt(self.model, self.optimizer)
            self.logger.end(summarize_vals=True)

class ImagenetTrainer(BaseTrainer):
    def _init_loaders(self):
        allreduce_batch_size = self.batch_size * self.batches_per_allreduce

        traindir = os.path.join(self.dataset_path, 'train')
        valdir = os.path.join(self.dataset_path, 'val')
        self.train_dataset = StridedImageFolder(
                traindir,
                transforms.Compose([
                        transforms.RandomResizedCrop(224),
                        transforms.RandomHorizontalFlip(),
                        transforms.ToTensor(),
                        self.normalize,]),
                stride=self.stride)
        self.val_dataset = StridedImageFolder(
                valdir,
                transforms.Compose([
                        transforms.Resize(256),
                        transforms.CenterCrop(224),
                        transforms.ToTensor(),
                        self.normalize,]),
                stride=self.stride)

        self.train_sampler = torch.utils.data.distributed.DistributedSampler(
                self.train_dataset, num_replicas=hvd.size(), rank=hvd.rank())
        self.val_sampler = torch.utils.data.distributed.DistributedSampler(
                self.val_dataset, num_replicas=hvd.size(), rank=hvd.rank())
        self.train_loader = torch.utils.data.DataLoader(
                self.train_dataset, batch_size=allreduce_batch_size,
                sampler=self.train_sampler, num_workers=8, pin_memory=True)
        self.val_loader = torch.utils.data.DataLoader(
                self.val_dataset, batch_size=allreduce_batch_size,
                sampler=self.val_sampler, num_workers=8, pin_memory=True,
                shuffle=False)

class CIFAR10Trainer(BaseTrainer):
    def __init__(self, **kwargs):
        if 'epochs' not in kwargs:
            kwargs['epochs'] = 200
        if 'base_lr' not in kwargs:
            kwargs['base_lr'] = 0.1
        super().__init__(**kwargs)

    def _adjust_learning_rate(self, epoch, batch_idx):
        if epoch < 100:
            lr = 0.1
        elif epoch < 150:
            lr = 0.01
        else:
            lr = 0.001
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr

    def _init_loaders(self):
        allreduce_batch_size = self.batch_size * self.batches_per_allreduce

        if hvd.local_rank() != 0:
            hvd.allreduce(torch.tensor(0), name='barrier')
        self.train_dataset = datasets.CIFAR10(
                root=self.dataset_path, download=(hvd.local_rank() == 0),
                train=True,
                transform=transforms.Compose([
                        transforms.RandomHorizontalFlip(),
                        transforms.RandomCrop(32, 4),
                        transforms.ToTensor(),
                        self.normalize,]))
        if hvd.local_rank() == 0:
            hvd.allreduce(torch.tensor(0), name='barrier')
        self.val_dataset = datasets.CIFAR10(
                root=self.dataset_path,
                train=False,
                transform=transforms.Compose([
                        transforms.ToTensor(),
                        self.normalize,]))
        self.train_loader = torch.utils.data.DataLoader(
                self.train_dataset, batch_size=allreduce_batch_size,
                shuffle=True, num_workers=8, pin_memory=True)
        self.val_loader = torch.utils.data.DataLoader(
                self.val_dataset, batch_size=allreduce_batch_size,
                shuffle=False, num_workers=8, pin_memory=True)



import os
import shutil
import tempfile

import torchvision

class StridedImageFolder(torchvision.datasets.ImageFolder):
    def __init__(self, root, *args, **kwargs):
        self.stride = kwargs['stride']
        del kwargs['stride']

        self.new_root = tempfile.mkdtemp()
        classes = [d for d in os.listdir(root) if os.path.isdir(os.path.join(root, d))]
        classes.sort()
        classes = classes[::self.stride]
        for cls in classes:
            os.symlink(os.path.join(root, cls), os.path.join(self.new_root, cls))

        super().__init__(self.new_root, *args, **kwargs)

    def __del__(self):
        shutil.rmtree(self.new_root)



import random

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from advex_uar.attacks.attacks import AttackWrapper
from advex_uar.attacks.snow import snow_creator, make_kernels

def apply_snow(img, snow, scale, discolor=0.25):    
    out = (1 - discolor) * img +\
    discolor * torch.max(img, (0.2126 * img[:,0:1] + 0.7152 * img[:,1:2] + 0.0722 * img[:,2:3]) * 1.5 + 0.5)   
    return torch.clamp(out + scale[:, None, None, None] * snow, 0, 1)

class SnowAttack(AttackWrapper):
    def __init__(self, nb_its, eps_max, step_size, resol, rand_init=True, scale_each=False,
                 budget=0.2):
        """
        Parameters:
            nb_its (int):          Number of GD iterations.
            eps_max (float):       The max norm, in pixel space.
            step_size (float):     The max step size, in pixel space.
            resol (int):           Side length of the image.
            rand_init (bool):      Whether to init randomly in the norm ball
            scale_each (bool):     Whether to scale eps for each image in a batch separately
            budget (float):        Controls rate parameter of snowflakes
        """
        super().__init__(resol)
        self.nb_its = nb_its
        self.eps_max = eps_max
        self.step_size = step_size
        self.resol = resol
        self.rand_init = rand_init
        self.scale_each = scale_each
        self.budget = budget

        self.criterion = nn.CrossEntropyLoss().cuda()
        self.nb_backward_steps = self.nb_its    

    def _init(self, batch_size):
        # flake intensities follow an exponential distribution
        flake_intensities = torch.exp(-1./(self.budget)*torch.rand(batch_size,7,self.resol//4,self.resol//4)).cuda()
        flake_intensities.requires_grad_(True)
        
        return flake_intensities

    def _forward(self, pixel_model, pixel_img, target, avoid_target=True, scale_eps=False):
        pixel_inp = pixel_img.detach()
        batch_size = pixel_img.size(0)
        
        if scale_eps:
            if self.scale_each:
                rand = torch.rand(pixel_img.size()[0], device='cuda')
            else:
                rand = random.random() * torch.ones(pixel_img.size()[0], device='cuda')
            base_eps = rand.mul(self.eps_max)
            step_size = self.step_size * torch.ones(pixel_img.size()[0], device='cuda')
        else:
            base_eps = self.eps_max * torch.ones(pixel_img.size()[0], device='cuda')
            step_size = self.step_size * torch.ones(pixel_img.size()[0], device='cuda')
        
        flake_intensities = self._init(batch_size)
        kernels = make_kernels()
        snow = snow_creator(flake_intensities, kernels, self.resol)
        s = pixel_model(apply_snow(pixel_inp / 255., snow, base_eps) * 255)

        for it in range(self.nb_its):
            loss = self.criterion(s, target)
            loss.backward()
            '''
            Because of batching, this grad is scaled down by 1 / batch_size, which does not matter
            for what follows because of normalization.
            '''            
            if avoid_target:
                # to avoid the target, we increase the loss
                grad = flake_intensities.grad.data
            else:
                # to hit the target, we reduce the loss
                grad = -flake_intensities.grad.data

            grad_sign = grad.sign()
            flake_intensities.data = flake_intensities.data + step_size[:, None, None, None] * grad_sign

            if it != self.nb_its - 1:
                snow = snow_creator(flake_intensities, kernels, self.resol)
                s = pixel_model(apply_snow(pixel_inp / 255., snow, base_eps) * 255)
                flake_intensities.grad.data.zero_()

                # keep intensities in valid range
                flake_intensities.detach()
                flake_intensities.data = flake_intensities.data.clamp(1e-9, 1)

                # ensure intensities do not exceed budget
                block_size = 8
                blocks = flake_intensities.size(-1)//block_size

                budget_per_region = F.adaptive_avg_pool2d(flake_intensities, blocks)
                budget_per_region[budget_per_region < self.budget] = self.budget

                for i in range(blocks):
                    for j in range(blocks):
                        flake_intensities.data[
                            :,:,i*block_size:(i+1)*block_size,
                            j*block_size:(j+1)*block_size] *= self.budget/budget_per_region[:,:,i,j].view(-1,7,1,1)

                flake_intensities.requires_grad_()

        snow = snow_creator(flake_intensities, kernels, self.resol)
        pixel_result = apply_snow(pixel_inp / 255., snow, base_eps) * 255
        return pixel_result




import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import itertools

# From https://gist.github.com/deanmark/9aec75b7dc9fa71c93c4bc85c5438777
def tensordot_pytorch(a, b, dims=2):
    axes=dims
    # code adapted from numpy
    try:
        iter(axes)
    except Exception:
        axes_a = list(range(-axes, 0))
        axes_b = list(range(0, axes))
    else:
        axes_a, axes_b = axes
    try:
        na = len(axes_a)
        axes_a = list(axes_a)
    except TypeError:
        axes_a = [axes_a]
        na = 1
    try:
        nb = len(axes_b)
        axes_b = list(axes_b)
    except TypeError:
        axes_b = [axes_b]
        nb = 1
    
    # uncomment in pytorch >= 0.5
    # a, b = torch.as_tensor(a), torch.as_tensor(b)
    as_ = a.shape
    nda = a.dim()
    bs = b.shape
    ndb = b.dim()
    equal = True
    if na != nb:
        equal = False
    else:
        for k in range(na):
            if as_[axes_a[k]] != bs[axes_b[k]]:
                equal = False
                break
            if axes_a[k] < 0:
                axes_a[k] += nda
            if axes_b[k] < 0:
                axes_b[k] += ndb
    if not equal:
        raise ValueError("shape-mismatch for sum")

    # Move the axes to sum over to the end of "a"
    # and to the front of "b"
    notin = [k for k in range(nda) if k not in axes_a]
    newaxes_a = notin + axes_a
    N2 = 1
    for axis in axes_a:
        N2 *= as_[axis]
    newshape_a = (int(np.multiply.reduce([as_[ax] for ax in notin])), N2)
    olda = [as_[axis] for axis in notin]

    notin = [k for k in range(ndb) if k not in axes_b]
    newaxes_b = axes_b + notin
    N2 = 1
    for axis in axes_b:
        N2 *= bs[axis]
    newshape_b = (N2, int(np.multiply.reduce([bs[ax] for ax in notin])))
    oldb = [bs[axis] for axis in notin]

    at = a.permute(newaxes_a).reshape(newshape_a)
    bt = b.permute(newaxes_b).reshape(newshape_b)

    res = at.matmul(bt)
    return res.reshape(olda + oldb)

# based on https://github.com/rshin/differentiable-jpeg/blob/master/jpeg.py
# Thanks to Dan Hendrycks for adapting the code to pytorch

def rgb_to_ycbcr(image):
  matrix = np.array(
      [[65.481, 128.553, 24.966],
       [-37.797, -74.203, 112.],
       [112., -93.786, -18.214]],
      dtype=np.float32).T / 255
  shift = torch.as_tensor([16., 128., 128.], device="cuda")

  # result = torch.tensordot(image, torch.as_tensor(matrix, device="cuda"), dims=1) + shift
  result = tensordot_pytorch(image, matrix, dims=1) + shift
  result.view(image.size())
  return result


def rgb_to_ycbcr_jpeg(image):
  matrix = np.array(
      [[0.299, 0.587, 0.114],
       [-0.168736, -0.331264, 0.5],
       [0.5, -0.418688, -0.081312]],
      dtype=np.float32).T
  shift = torch.as_tensor([0., 128., 128.], device="cuda")

  # result = torch.tensordot(image, torch.as_tensor(matrix, device="cuda"), dims=1) + shift
  result = tensordot_pytorch(image, torch.as_tensor(matrix, device='cuda'), dims=1) + shift
  result.view(image.size())
  return result


# 2. Chroma subsampling
def downsampling_420(image):
  # input: batch x height x width x 3
  # output: tuple of length 3
  #   y:  batch x height x width
  #   cb: batch x height/2 x width/2
  #   cr: batch x height/2 x width/2
  y, cb, cr = image[...,0], image[...,1], image[...,2]
  # requires that height and width are divisible by 2 to avoid
  # padding issues
  cb = F.avg_pool2d(cb, kernel_size=2)
  cr = F.avg_pool2d(cr, kernel_size=2)
  return y, cb, cr
  # return (y.squeeze(), cb.squeeze(), cr.squeeze())


# 3. Block splitting
# From https://stackoverflow.com/questions/41564321/split-image-tensor-into-small-patches
def image_to_patches(image):
  # input: batch x h x w
  # output: batch x h*w/64 x h x w
  k = 8
  batch_size, height, width = image.size()
  image_reshaped = image.view(batch_size, height // k, k, -1, k)
  image_transposed = torch.transpose(image_reshaped, 2, 3)
  return image_transposed.contiguous().view(batch_size, -1, k, k)


# 4. DCT
def dct_8x8_ref(image):
  image = image - 128
  result = np.zeros((8, 8), dtype=np.float32)
  for u, v in itertools.product(range(8), range(8)):
    value = 0
    for x, y in itertools.product(range(8), range(8)):
      value += image[x, y] * np.cos((2 * x + 1) * u * np.pi / 16) * np.cos(
          (2 * y + 1) * v * np.pi / 16)
    result[u, v] = value
  alpha = np.array([1. / np.sqrt(2)] + [1] * 7)
  scale = np.outer(alpha, alpha) * 0.25
  return result * scale


def dct_8x8(image):
  image = image - 128
  tensor = np.zeros((8, 8, 8, 8), dtype=np.float32)
  for x, y, u, v in itertools.product(range(8), repeat=4):
    tensor[x, y, u, v] = np.cos((2 * x + 1) * u * np.pi / 16) * np.cos(
        (2 * y + 1) * v * np.pi / 16)
  alpha = np.array([1. / np.sqrt(2)] + [1] * 7)
  scale = torch.FloatTensor(np.outer(alpha, alpha) * 0.25).cuda()
  #result = scale * torch.tensordot(image, torch.as_tensor(tensor, device="cuda"), dims=2)
  result = scale * tensordot_pytorch(image, torch.as_tensor(tensor, device="cuda"), dims=2)
  result.view(image.size())
  return result


def make_quantization_tables(self):
    # 5. Quantizaztion
    self.y_table = torch.as_tensor(np.array(
        [[16, 11, 10, 16, 24, 40, 51, 61],
         [12, 12, 14, 19, 26, 58, 60, 55],
         [14, 13, 16, 24, 40, 57, 69, 56],
         [14, 17, 22, 29, 51, 87, 80, 62],
         [18, 22, 37, 56, 68, 109, 103, 77],
         [24, 35, 55, 64, 81, 104, 113, 92],
         [49, 64, 78, 87, 103, 121, 120, 101],
         [72, 92, 95, 98, 112, 100, 103, 99]],
        dtype=np.float32).T, device="cuda")
    c_table = np.empty((8, 8), dtype=np.float32)
    c_table.fill(99)
    c_table[:4, :4] = np.array([[17, 18, 24, 47], [18, 21, 26, 66],
                                [24, 26, 56, 99], [47, 66, 99, 99]]).T
    self.c_table = torch.as_tensor(c_table, device="cuda")


def y_quantize(self, image, rounding, rounding_var, factor=1):
  image = image / (self.y_table * factor)
  image = rounding(image, rounding_var)
  return image


def c_quantize(self, image, rounding, rounding_var, factor=1):
  image = image / (self.c_table * factor)
  image = rounding(image, rounding_var)
  return image


# -5. Dequantization
def y_dequantize(self, image, factor=1):
  return image * (self.y_table * factor)


def c_dequantize(self, image, factor=1):
  return image * (self.c_table * factor)


# -4. Inverse DCT
def idct_8x8_ref(image):
  alpha = np.array([1. / np.sqrt(2)] + [1] * 7)
  alpha = np.outer(alpha, alpha)
  image = image * alpha

  result = np.zeros((8, 8), dtype=np.float32)
  for u, v in itertools.product(range(8), range(8)):
    value = 0
    for x, y in itertools.product(range(8), range(8)):
      value += image[x, y] * np.cos((2 * u + 1) * x * np.pi / 16) * np.cos(
          (2 * v + 1) * y * np.pi / 16)
    result[u, v] = value
  return result * 0.25 + 128


def idct_8x8(image):
  alpha = np.array([1. / np.sqrt(2)] + [1] * 7)
  alpha = torch.FloatTensor(np.outer(alpha, alpha)).cuda()
  image = image * alpha

  tensor = np.zeros((8, 8, 8, 8), dtype=np.float32)
  for x, y, u, v in itertools.product(range(8), repeat=4):
    tensor[x, y, u, v] = np.cos((2 * u + 1) * x * np.pi / 16) * np.cos(
        (2 * v + 1) * y * np.pi / 16)
  # result = 0.25 * torch.tensordot(image, torch.as_tensor(tensor, device="cuda"), dims=2) + 128
  result = 0.25 * tensordot_pytorch(image, torch.as_tensor(tensor, device="cuda"), dims=2) + 128
  result.view(image.size())
  return result


# -3. Block joining
def patches_to_image(patches, height, width):
  # input: batch x h*w/64 x h x w
  # output: batch x h x w
  height = int(height)
  width = int(width)
  k = 8
  batch_size = patches.size(0)
  image_reshaped = patches.view(batch_size, height // k, width // k, k, k)
  image_transposed = torch.transpose(image_reshaped, 2, 3)
  return image_transposed.contiguous().view(batch_size, height, width)

# -2. Chroma upsampling
def upsampling_420(y, cb, cr):
  # input:
  #   y:  batch x height x width
  #   cb: batch x height/2 x width/2
  #   cr: batch x height/2 x width/2
  # output:
  #   image: batch x height x width x 3
  def repeat(x, k=2):
    height, width = x.size()[1:3]
    x = x.unsqueeze(-1)
    x = x.repeat((1, 1, k, k))
    x = x.view(-1, height * k, width * k)
    return x

  cb = repeat(cb)
  cr = repeat(cr)
  return torch.stack((y, cb, cr), dim=-1)


# -1. YCbCr -> RGB
def ycbcr_to_rgb(image):
  matrix = np.array(
      [[298.082, 0, 408.583],
       [298.082, -100.291, -208.120],
       [298.082, 516.412, 0]],
      dtype=np.float32).T / 256
  shift = torch.as_tensor([-222.921, 135.576, -276.836], device="cuda")

  # result = torch.tensordot(image, torch.tensor(matrix, device="cuda"), dims=1) + shift
  result = tensordot_pytorch(image, torch.tensor(matrix, device="cuda"), dims=1) + shift
  result.view(image.size())
  return result


def ycbcr_to_rgb_jpeg(image):
  matrix = np.array(
      [[1., 0., 1.402],
       [1, -0.344136, -0.714136],
       [1, 1.772, 0]],
      dtype=np.float32).T
  shift = torch.FloatTensor([0, -128, -128]).cuda()

  # result = torch.tensordot(image + shift, torch.tensor(matrix, device="cuda"), dims=1)
  result = tensordot_pytorch(image + shift, torch.tensor(matrix, device="cuda"), dims=1)
  result.view(image.size())
  return result


def jpeg_compress_decode(self, image_channels_first, rounding_vars, lambder, downsample_c=True,
                         factor=1):
  def noisy_round(x, noise):
    return x + lambder[:, None, None, None] * (noise - 0.5)
  
  #image = torch.as_tensor(image)
  image = torch.transpose(image_channels_first, 1, 3)
  height, width = image.size()[1:3]

  orig_height, orig_width = height, width
  if height % 16 != 0 or width % 16 != 0:
    # Round up to next multiple of 16
    height = ((height - 1) // 16 + 1) * 16
    width = ((width - 1) // 16 + 1) * 16

    vpad = height - orig_height
    wpad = width - orig_width
    top = vpad // 2
    bottom = vpad - top
    left = wpad // 2
    right = wpad - left

    # image = tf.pad(image, [[0, 0], [top, bottom], [left, right], [0, 0]], 'SYMMETRIC')
    # image = tf.pad(image, [[0, 0], [0, vpad], [0, wpad], [0, 0]], 'SYMMETRIC')
    # image = F.pad(image, (vpad, wpad), 'replicate')
    image = F.pad(image, (left, right, top, bottom), 'replicate')

  # "Compression"
  image = rgb_to_ycbcr_jpeg(image)
  if downsample_c:
    y, cb, cr = downsampling_420(image)
  else:
    y, cb, cr = torch.split(image, 3, dim=3)
  components = {'y': y, 'cb': cb, 'cr': cr}
  for k in components.keys():
    comp = components[k]
    comp = image_to_patches(comp)
    comp = dct_8x8(comp)
    if k == 'y':
        comp = y_quantize(self, comp, noisy_round, 0.5 + 0.5 * rounding_vars[0], factor)
    elif k  == 'cb':
        comp = c_quantize(self, comp, noisy_round, 0.5 + 0.5 * rounding_vars[1], factor)
    else:
        comp = c_quantize(self, comp, noisy_round, 0.5 + 0.5 * rounding_vars[2], factor)
    components[k] = comp

  # Decoding
  for k in components.keys():
    comp = components[k]
    comp = c_dequantize(self, comp, factor) if k in ('cb', 'cr') else y_dequantize(
        self, comp, factor)
    comp = idct_8x8(comp)
    if k in ('cb', 'cr'):
      if downsample_c:
        comp = patches_to_image(comp, height / 2, width / 2)
      else:
        comp = patches_to_image(comp, height, width)
    else:
      comp = patches_to_image(comp, height, width)
    components[k] = comp

  y, cb, cr = components['y'], components['cb'], components['cr']
  if downsample_c:
    image = upsampling_420(y, cb, cr)
  else:
    image = torch.stack((y, cb, cr), dim=-1)
  image = ycbcr_to_rgb_jpeg(image)

  # Crop to original size
  if orig_height != height or orig_width != width:
    #image = image[:, top:-bottom, left:-right]
    image = image[:, :-vpad, :-wpad]

  # Hack: RGB -> YUV -> RGB sometimes results in incorrect values
  #    min_value = tf.minimum(tf.reduce_min(image), 0.)
  #    max_value = tf.maximum(tf.reduce_max(image), 255.)
  #    value_range = max_value - min_value
  #    image = 255 * (image - min_value) / value_range
  image = torch.clamp(image, 0, 255)

  return torch.transpose(image, 1, 3)

def quality_to_factor(quality):
    if quality < 50:
        return 50./quality
    else:
        return (200. - quality * 2)/100.


class JPEG(nn.Module):
    def __init__(self):
        super(JPEG, self).__init__()
        make_quantization_tables(self)

    def forward(self, pixel_inp, rounding_vars, epsilon):
        return jpeg_compress_decode(self, pixel_inp, rounding_vars, epsilon)
        




import random

import numpy as np
import torch
import torch.nn as nn
import scipy.sparse as sparse

from advex_uar.attacks.attacks import AttackWrapper
from advex_uar.attacks.gabor import get_gabor_with_sides, valid_position, gabor_rand_distributed

class GaborAttack(AttackWrapper):
    def __init__(self, nb_its, eps_max, step_size, resol, rand_init=True, scale_each=False):
        """
        Parameters:
            nb_its (int):          Number of GD iterations.
            eps_max (float):       The max norm, in pixel space.
            step_size (float):     The max step size, in pixel space.
            resol (int):           Side length of the image.
            rand_init (bool):      Whether to init randomly in the norm ball
            scale_each (bool):     Whether to scale eps for each image in a batch separately
        """
        super().__init__(resol)
        self.nb_its = nb_its
        self.eps_max = eps_max
        self.step_size = step_size
        self.resol = resol
        self.rand_init = rand_init
        self.scale_each = scale_each

        self.criterion = nn.CrossEntropyLoss().cuda()
        self.nb_backward_steps = self.nb_its

    def _init(self, batch_size, num_kern):
        grid = 14

        if self.rand_init:
            sparse_matrices = []
            sp_conv_numpy = sparse.random(self.resol*batch_size, self.resol,
                            density= 1. / grid, format='csr')
            sp_conv_numpy.data = sp_conv_numpy.data * 2 - 1
            sp_conv = torch.FloatTensor(sp_conv_numpy.todense()).view(
                        batch_size, self.resol, self.resol)

            mask = (sp_conv == 0).cuda().float().view(-1, 1, self.resol, self.resol)
            gabor_vars = sp_conv.clone().cuda().view(-1, 1, self.resol, self.resol)
            gabor_vars.requires_grad_(True)
        return gabor_vars, mask

    def _get_gabor_kernel(self, batch_size):
        # make gabor filters to convolve with variables
        k_size = 23
        kernels = []
        for b in range(batch_size):
            sides = np.random.randint(10) + 1
            sigma = 0.3 * torch.rand(1) +  0.1
            Lambda = (k_size / 4. - 3) * torch.rand(1) + 3
            theta = np.pi * torch.rand(1)

            kernels.append(get_gabor_with_sides(k_size, sigma, Lambda, theta, sides).cuda())
        gabor_kernel = torch.cat(kernels, 0).view(-1, 1, k_size, k_size)
        return gabor_kernel

    def _forward(self, pixel_model, pixel_img, target, avoid_target=True, scale_eps=False):
        pixel_inp = pixel_img.detach()
        batch_size = pixel_img.size(0)

        if scale_eps:
            if self.scale_each:
                rand = torch.rand(pixel_img.size()[0], device='cuda')
            else:
                rand = random.random() * torch.ones(pixel_img.size()[0], device='cuda')
            base_eps = rand.mul(self.eps_max)
            step_size = self.step_size * torch.ones(pixel_img.size()[0], device='cuda')
        else:
            base_eps = self.eps_max * torch.ones(pixel_img.size()[0], device='cuda')
            step_size = self.step_size * torch.ones(pixel_img.size()[0], device='cuda')

        gabor_kernel = self._get_gabor_kernel(batch_size)
        num_kern = np.random.randint(50) + 1
        gabor_vars, mask = self._init(batch_size, num_kern)
        gabor_noise = gabor_rand_distributed(gabor_vars, gabor_kernel)
        gabor_noise = gabor_noise.expand(-1, 3, -1, -1)
        s = pixel_model(torch.clamp(pixel_inp + base_eps[:, None, None, None] * gabor_noise, 0., 255.))
        for it in range(self.nb_its):
            loss = self.criterion(s, target)
            loss.backward()
            '''
            Because of batching, this grad is scaled down by 1 / batch_size, which does not matter
            for what follows because of normalization.
            '''
            if avoid_target:
                # to avoid the target, we increase the loss
                grad = gabor_vars.grad.data
            else:
                # to hit the target, we reduce the loss
                grad = -gabor_vars.grad.data

            grad_sign = grad.sign()
            gabor_vars.data = gabor_vars.data + step_size[:, None, None, None] * grad_sign
            gabor_vars.data = torch.clamp(gabor_vars.data, -1, 1) * mask

            if it != self.nb_its - 1:
                gabor_noise = gabor_rand_distributed(gabor_vars, gabor_kernel).expand(-1, 3, -1, -1)
                s = pixel_model(torch.clamp(pixel_inp + base_eps[:, None, None, None] * gabor_noise, 0., 255.))
                gabor_vars.grad.data.zero_()
        pixel_result = torch.clamp(pixel_inp + base_eps[:, None, None, None] * gabor_rand_distributed(gabor_vars, gabor_kernel), 0., 255.)
        return pixel_result



import random

import numpy as np
import torch
import torch.nn as nn

from advex_uar.attacks.attacks import AttackWrapper

class PGDAttack(AttackWrapper):
    def __init__(self, nb_its, eps_max, step_size, resol, norm='linf', rand_init=True, scale_each=False):
        """
        Parameters:
            nb_its (int):          Number of PGD iterations.
            eps_max (float):       The max norm, in pixel space.
            step_size (float):     The max step size, in pixel space.
            resol (int):           Side length of the image.
            norm (str):            Either 'linf' or 'l2'
            rand_init (bool):      Whether to init randomly in the norm ball
            scale_each (bool):     Whether to scale eps for each image in a batch separately
        """
        super().__init__(resol)
        self.nb_its = nb_its
        self.eps_max = eps_max
        self.step_size = step_size
        self.resol = resol
        self.norm = norm
        self.rand_init = rand_init
        self.scale_each = scale_each

        self.criterion = nn.CrossEntropyLoss().cuda()
        self.nb_backward_steps = self.nb_its
        
    def _run_one(self, pixel_model, pixel_inp, delta, target, eps, step_size, avoid_target=True):
        s = pixel_model(pixel_inp + delta)
        if self.norm == 'l2':
            l2_max = eps
        for it in range(self.nb_its):
            loss = self.criterion(s, target)
            loss.backward()
            '''
            Because of batching, this grad is scaled down by 1 / batch_size, which does not matter
            for what follows because of normalization.
            '''
            if avoid_target:
                # to avoid the target, we increase the loss
                grad = delta.grad.data
            else:
                # to hit the target, we reduce the loss
                grad = -delta.grad.data

            if self.norm == 'linf':
                grad_sign = grad.sign()
                delta.data = delta.data + step_size[:, None, None, None] * grad_sign
                delta.data = torch.max(torch.min(delta.data, eps[:, None, None, None]), -eps[:, None, None, None])
                delta.data = torch.clamp(pixel_inp.data + delta.data, 0., 255.) - pixel_inp.data
            elif self.norm == 'l2':
                batch_size = delta.data.size()[0]
                grad_norm = torch.norm(grad.view(batch_size, -1), 2.0, dim=1)
                normalized_grad = grad / grad_norm[:, None, None, None]                
                delta.data = delta.data + step_size[:, None, None, None] * normalized_grad
                l2_delta = torch.norm(delta.data.view(batch_size, -1), 2.0, dim=1)
                # Check for numerical instability
                proj_scale = torch.min(torch.ones_like(l2_delta, device='cuda'), l2_max / l2_delta)
                delta.data *= proj_scale[:, None, None, None]
                delta.data = torch.clamp(pixel_inp.data + delta.data, 0., 255.) - pixel_inp.data
            else:
                raise NotImplementedError

            if it != self.nb_its - 1:
                s = pixel_model(pixel_inp + delta)
                delta.grad.data.zero_()
        return delta

    def _init(self, shape, eps):
        if self.rand_init:
            if self.norm == 'linf':
                init = torch.rand(shape, dtype=torch.float32, device='cuda') * 2 - 1
            elif self.norm == 'l2':
                init = torch.randn(shape, dtype=torch.float32, device='cuda')
                init_norm = torch.norm(init.view(init.size()[0], -1), 2.0, dim=1)
                normalized_init = init / init_norm[:, None, None, None]
                dim = init.size()[1] * init.size()[2] * init.size()[3]
                rand_norms = torch.pow(torch.rand(init.size()[0], dtype=torch.float32, device='cuda'), 1/dim)
                init = normalized_init * rand_norms[:, None, None, None]
            else:
                raise NotImplementedError
            init = eps[:, None, None, None] * init
            init.requires_grad_()
            return init
        else:
            return torch.zeros(shape, requires_grad=True, device='cuda')
    
    def _forward(self, pixel_model, pixel_img, target, avoid_target=True, scale_eps=False):
        if scale_eps:
            if self.scale_each:
                rand = torch.rand(pixel_img.size()[0], device='cuda')
            else:
                rand = random.random() * torch.ones(pixel_img.size()[0], device='cuda')
            base_eps = rand.mul(self.eps_max)
            step_size = rand.mul(self.step_size)
        else:
            base_eps = self.eps_max * torch.ones(pixel_img.size()[0], device='cuda')
            step_size = self.step_size * torch.ones(pixel_img.size()[0], device='cuda')

        pixel_inp = pixel_img.detach()
        pixel_inp.requires_grad = True
        delta = self._init(pixel_inp.size(), base_eps)
        if self.nb_its > 0:
            delta = self._run_one(pixel_model, pixel_inp, delta, target, base_eps,
                                  step_size, avoid_target=avoid_target)
        else:
            delta.data = torch.clamp(pixel_inp.data + delta.data, 0., 255.) - pixel_inp.data
        pixel_result = pixel_inp + delta
        return pixel_result




import argparse
import itertools
import math
import numbers
import os
import pickle
import sys

import numpy as np
import torch
import torch.nn.functional as F

def valid_position(size, x, y):
    if x < 0 or x >= size: return False
    if y < 0 or y >= size: return False
    return True

def normalize(orig):
    batch_size = orig.size(0)
    omax = torch.max(orig.view(batch_size, -1), 1)[0].detach().view(batch_size,1,1,1)
    omin  = torch.min(orig.view(batch_size, -1), 1)[0].detach().view(batch_size,1,1,1)
    return (orig - omin) / (omax - omin)

def get_gabor(k_size, sigma, Lambda, theta):
    y, x = torch.meshgrid([torch.linspace(-0.5, 0.5, k_size), torch.linspace(-0.5, 0.5, k_size)])
    rotx = x * torch.cos(theta) + y * torch.sin(theta)
    roty = -x * torch.sin(theta) + y * torch.cos(theta)
    g = torch.zeros(y.shape)
    g = torch.exp(-0.5 * (rotx ** 2 / (sigma + 1e-3) ** 2 + roty ** 2 / (sigma + 1e-3) ** 2))
    g = g * torch.cos(2 * np.pi * Lambda * rotx)
    return g

def get_gabor_with_sides(k_size, sigma, Lambda, theta, sides=3):
    g = get_gabor(k_size, sigma, Lambda, theta)
    for i in range(1, sides):
        g += get_gabor(k_size, sigma, Lambda, theta + np.pi * i / sides)
    return g

def normalize_var(orig):
    batch_size = orig.size(0)

    # Spectral variance
    mean = torch.mean(orig.view(batch_size, -1), 1).view(batch_size, 1, 1, 1)
    spec_var = torch.rfft(torch.pow(orig -  mean, 2), 2)

    # Normalization
    imC = torch.sqrt(torch.irfft(spec_var, 2, signal_sizes=orig.size()[2:]).abs())
    imC /= torch.max(imC.view(batch_size, -1), 1)[0].view(batch_size, 1, 1, 1)
    minC = 0.001
    imK =  (minC + 1) / (minC + imC)

    mean, imK = mean.detach(), imK.detach()
    img = mean + (orig -  mean) * imK
    return normalize(img)

def gabor_rand_distributed(sp_conv, gabor_kernel):
    # code converted from https://github.com/kenny-co/procedural-advml

    batch_size = sp_conv.size(0)
    # reshape batch dimension to channel dimension to use group convolution
    # so that data processes in parallel
    sp_conv = sp_conv.view(1, batch_size, sp_conv.size(-2), sp_conv.size(-1))
    sp_conv = F.conv2d(sp_conv, weight=gabor_kernel, stride=1, groups=batch_size, padding=11)
    sp_conv = sp_conv.view(batch_size, 1, sp_conv.size(-2), sp_conv.size(-1))

    return normalize_var(sp_conv)



import random

import torch
import torch.nn as nn

from advex_uar.attacks.attacks import AttackWrapper
from advex_uar.attacks.jpeg import JPEG

class JPEGAttack(AttackWrapper):
    def __init__(self, nb_its, eps_max, step_size, resol,
                 rand_init=True, opt='linf', scale_each=False, l1_max=2.):
        '''
        Arguments:
            nb_its (int):          Number of iterations
            eps_max (float):       Maximum flow, in L_inf norm, in pixels
            step_size (float):     Maximum step size in L_inf norm, in pixels
            resol (int):           Side length of images, in pixels
            rand_init (bool):      Whether to do a random init
            opt (string):          Which optimization algorithm to use, either 'linf', 'l1', or 'l2'
            scale_each (bool):     Whether to scale eps for each image in a batch separately
        '''
        super().__init__(resol)
        self.nb_its = nb_its
        self.eps_max = eps_max
        self.step_size = step_size
        self.rand_init = rand_init
        self.opt = opt
        if opt not in ['linf', 'l1', 'l2']:
            raise NotImplementedError
        self.scale_each = scale_each
        self.l1_max = l1_max
        
        self.criterion = nn.CrossEntropyLoss().cuda()
        self.nb_backward_steps = nb_its
        self.jpeg = JPEG().cuda()

    def _convert_cat_var(self, cat_var, batch_size, height, width):
        y_var = cat_var[:, :height//8 * width//8 * 8 * 8].view((batch_size, height//8 * width//8, 8, 8))
        cb_var = cat_var[:, height//8 * width//8 * 8 * 8: height//8 * width//8 * 8 * 8 + height//16 * width//16 * 8 * 8].view((batch_size, height//16 * width//16, 8, 8))
        cr_var = cat_var[:, height//8 * width//8 * 8 * 8 + height//16 * width//16 * 8 * 8: height//8 * width//8 * 8 * 8 + 2 * height//16 * width//16 * 8 * 8].view((batch_size, height//16 * width//16, 8, 8))
        return y_var, cb_var, cr_var
    
    def _jpeg_cat(self, pixel_inp, cat_var, base_eps, batch_size, height, width):
        y_var, cb_var, cr_var = self._convert_cat_var(cat_var, batch_size, height, width)
        return self.jpeg(pixel_inp, [y_var, cb_var, cr_var], base_eps)
        
    def _run_one_pgd(self, pixel_model, pixel_inp, cat_var, target, base_eps, step_size, avoid_target=True):
        batch_size, channels, height, width = pixel_inp.size()
        pixel_inp_jpeg = self._jpeg_cat(pixel_inp, cat_var, base_eps, batch_size, height, width)
        s = pixel_model(pixel_inp_jpeg)

        for it in range(self.nb_its):
            loss = self.criterion(s, target)
            loss.backward()

            if avoid_target:
                grad = cat_var.grad.data
            else:
                grad = -cat_var.grad.data

            if self.opt == 'linf':
                grad_sign = grad.sign()
                cat_var.data = cat_var.data + step_size[:, None] * grad_sign
                cat_var.data = torch.max(torch.min(cat_var.data, base_eps[:, None]),
                                         -base_eps[:, None]) # cannot use torch.clamp for tensors
            elif self.opt == 'l2':
                batch_size = pixel_inp.size()[0]
                grad_norm = torch.norm(grad.view(batch_size, -1), 2.0, dim=1)
                normalized_grad = grad / grad_norm[:, None]                
                cat_var.data = cat_var.data + step_size[:, None] * normalized_grad
                l2_delta = torch.norm(cat_var.data.view(batch_size, -1), 2.0, dim=1)
                proj_scale = torch.min(torch.ones_like(l2_delta, device='cuda'), base_eps / l2_delta)
                cat_var.data *= proj_scale[:, None]
                cat_var.data = torch.clamp(cat_var.data, -self.l1_max, self.l1_max)
                    
            if it != self.nb_its - 1:
                # self.jpeg scales rounding_vars by base_eps, so we divide to rescale
                # its coordinates to [-1, 1]
                cat_var_temp = cat_var / base_eps[:, None]
                pixel_inp_jpeg = self._jpeg_cat(pixel_inp, cat_var_temp, base_eps, batch_size, height, width)
                s = pixel_model(pixel_inp_jpeg)
            cat_var.grad.data.zero_()
        return cat_var
        
    def _run_one_fw(self, pixel_model, pixel_inp, cat_var, target, base_eps, avoid_target=True):
        batch_size, channels, height, width = pixel_inp.size()
        pixel_inp_jpeg = self._jpeg_cat(pixel_inp, cat_var, base_eps, batch_size, height, width)
        s = pixel_model(pixel_inp_jpeg)

        for it in range(self.nb_its):
            loss = self.criterion(s, target)
            loss.backward()

            if avoid_target:
                grad = cat_var.grad.data
            else:
                grad = -cat_var.grad.data

            def where_float(cond, if_true, if_false):
                return cond.float() * if_true + (1-cond.float()) * if_false

            def where_long(cond, if_true, if_false):
                return cond.long() * if_true + (1-cond.long()) * if_false

            abs_grad = torch.abs(grad).view(batch_size, -1)
            num_pixels = abs_grad.size()[1]
            sign_grad = torch.sign(grad)

            bound = where_float(sign_grad > 0, self.l1_max - cat_var, cat_var + self.l1_max).view(batch_size, -1)
                
            k_min = torch.zeros((batch_size,1), dtype=torch.long, requires_grad=False, device='cuda')
            k_max = torch.ones((batch_size,1), dtype=torch.long, requires_grad=False, device='cuda') * num_pixels
                
            # cum_bnd[k] is meant to track the L1 norm we end up with if we take 
            # the k indices with the largest gradient magnitude and push them to their boundary values (0 or 255)
            values, indices = torch.sort(abs_grad, descending=True)
            bnd = torch.gather(bound, 1, indices)
            # subtract bnd because we don't want the cumsum to include the final element
            cum_bnd = torch.cumsum(bnd, 1) - bnd
                
            # this is hard-coded as floor(log_2(256 * 256 * 3))
            for _ in range(17):
                k_mid = (k_min + k_max) // 2
                l1norms = torch.gather(cum_bnd, 1, k_mid)
                k_min = where_long(l1norms > base_eps, k_min, k_mid)
                k_max = where_long(l1norms > base_eps, k_mid, k_max)
                
            # next want to set the gradient of indices[0:k_min] to their corresponding bound
            magnitudes = torch.zeros((batch_size, num_pixels), requires_grad=False, device='cuda')
            for bi in range(batch_size):
                magnitudes[bi, indices[bi, :k_min[bi,0]]] = bnd[bi, :k_min[bi,0]]
                magnitudes[bi, indices[bi, k_min[bi,0]]] = base_eps[bi] - cum_bnd[bi, k_min[bi,0]]
                
            delta_it = sign_grad * magnitudes.view(cat_var.size())
            # These should always be exactly epsilon
            # l1_check = torch.norm(delta_it.view(batch_size, -1), 1.0, dim=1) / num_pixels
            # print('l1_check: %s' % l1_check)
            cat_var.data = cat_var.data + (delta_it - cat_var.data) / (it + 1.0)

            if it != self.nb_its - 1:
                # self.jpeg scales rounding_vars by base_eps, so we divide to rescale
                # its coordinates to [-1, 1]
                cat_var_temp = cat_var / base_eps[:, None]
                pixel_inp_jpeg = self._jpeg_cat(pixel_inp, cat_var_temp, base_eps, batch_size, height, width)
                s = pixel_model(pixel_inp_jpeg)
            cat_var.grad.data.zero_()
        return cat_var

    def _init_empty(self, batch_size, height, width):
        shape = (batch_size, (height//8 * width//8 + height//16 * width//16 + height//16 * width//16) * 8 * 8)
        return torch.zeros(shape, device='cuda')
    
    def _init_linf(self, batch_size, height, width):
        shape = (batch_size, (height//8 * width//8 + height//16 * width//16 + height//16 * width//16) * 8 * 8)
        return torch.rand(shape, device='cuda') * 2 - 1
        
    def _init_l1(self, batch_size, height, width):
        # returns random initialization with L_1 norm of 1
        # algorithm from https://arxiv.org/abs/math/0503650, Theorem 1
        shape = (batch_size, (height//8 * width//8 + height//16 * width//16 + height//16 * width//16) * 8 * 8)
        exp = torch.empty(shape, dtype=torch.float32, device='cuda')
        exp.exponential_()
        signs = torch.sign(torch.randn(shape, dtype=torch.float32, device='cuda'))
        exp = exp * signs
        exp_y = torch.empty(shape[0], dtype=torch.float32, device='cuda')
        exp_y.exponential_()
        norm = exp_y + torch.norm(exp.view(shape[0], -1), 1.0, dim=1)
        init = exp / norm[:, None]
        return init

    def _init_l2(self, batch_size, height, width):
        shape = (batch_size, (height//8 * width//8 + height//16 * width//16 + height//16 * width//16) * 8 * 8)
        init = torch.randn(shape, dtype=torch.float32, device='cuda')
        init_norm = torch.norm(init.view(batch_size, -1), 2.0, dim=1)
        normalized_init = init / init_norm[:, None]
        rand_norms = torch.pow(torch.rand(init.size()[0], dtype=torch.float32, device='cuda'), 1/shape[1])
        init = normalized_init * rand_norms[:, None]
        return init
        
    def _init(self, batch_size, height, width, eps):
        if self.rand_init:
            if self.opt == 'linf':
                cat_var = self._init_linf(batch_size, height, width)
            elif self.opt == 'l1':
                cat_var = self._init_l1(batch_size, height, width)
            elif self.opt == 'l2':
                cat_var = self._init_l2(batch_size, height, width)
            else:
                raise NotImplementedError
        else:
            cat_var = self._init_empty(batch_size, height, width)
        cat_var.mul_(eps[:, None])
        cat_var.requires_grad_()
        return cat_var

    def _forward(self, pixel_model, pixel_img, target, scale_eps=False, avoid_target=False):
        if scale_eps:
            if self.scale_each:
                rand = torch.rand(pixel_img.size()[0], device='cuda')
            else:
                rand = random.random() * torch.ones(pixel_img.size()[0], device='cuda')
            base_eps = rand * self.eps_max
            step_size = rand * self.step_size
        else:
            base_eps = self.eps_max * torch.ones(pixel_img.size()[0], device='cuda')
            step_size = self.step_size * torch.ones(pixel_img.size()[0], device='cuda')
        batch_size, channels, height, width = pixel_img.size()
        if height % 16 != 0 or width % 16 != 0:
            raise Exception
        pixel_inp = pixel_img.detach()
        pixel_inp.requires_grad = True
        cat_var = self._init(batch_size, height, width, base_eps)
        if self.nb_its > 0:
            if self.opt in ['linf', 'l2']:
                cat_var = self._run_one_pgd(pixel_model, pixel_inp, cat_var, target,
                                            base_eps, step_size, avoid_target=avoid_target)
            elif self.opt == 'l1':
                cat_var = self._run_one_fw(pixel_model, pixel_inp, cat_var, target,
                                           base_eps, avoid_target=avoid_target)
            else:
                raise NotImplementedError
        # self.jpeg scales rounding_vars by base_eps, so we divide to rescale
        # its coordinates to [-1, 1]
        cat_var_temp = cat_var / base_eps[:, None]
        pixel_result = self._jpeg_cat(pixel_inp, cat_var_temp, base_eps, batch_size, height, width)
        return pixel_result




import argparse
import itertools
import math
import numbers
import os
import pickle
import sys

import numpy as np
import torch
import torch.nn.functional as F

def fog_creator(fog_vars, bsize=1, mapsize=256, wibbledecay=1.75):
    assert (mapsize & (mapsize - 1) == 0)
    maparray = torch.from_numpy(np.empty((bsize, mapsize, mapsize), dtype=np.float32)).cuda()
    maparray[:, 0, 0] = 0
    stepsize = mapsize
    wibble = 100
    
    var_num = 0
    
    def wibbledmean(array, var_num):
        result = array / 4. + fog_vars[var_num] * 2 * wibble - wibble
        return result
    
    def fillsquares(var_num):
        """For each square of points stepsize apart,
           calculate middle value as mean of points + wibble"""
        cornerref = maparray[:, 0:mapsize:stepsize, 0:mapsize:stepsize]
        squareaccum = cornerref + torch.roll(cornerref, -1, 1)
        squareaccum = squareaccum + torch.roll(squareaccum, -1, 2)
        maparray[:, stepsize // 2:mapsize:stepsize,
        stepsize // 2:mapsize:stepsize] = wibbledmean(squareaccum, var_num)
        return var_num + 1

    def filldiamonds(var_num):
        """For each diamond of points stepsize apart,
           calculate middle value as mean of points + wibble"""
        mapsize = maparray.size(1)
        drgrid = maparray[:, stepsize // 2:mapsize:stepsize, stepsize // 2:mapsize:stepsize]
        ulgrid = maparray[:, 0:mapsize:stepsize, 0:mapsize:stepsize]
        ldrsum = drgrid + torch.roll(drgrid, 2, 1)
        lulsum = ulgrid + torch.roll(ulgrid, -1, 2)
        ltsum = ldrsum + lulsum
        maparray[:, 0:mapsize:stepsize, stepsize // 2:mapsize:stepsize] = wibbledmean(ltsum, var_num)
        var_num += 1
        tdrsum = drgrid + torch.roll(drgrid, 2, 2)
        tulsum = ulgrid + torch.roll(ulgrid, -1, 1)
        ttsum = tdrsum + tulsum
        maparray[:, stepsize // 2:mapsize:stepsize, 0:mapsize:stepsize] = wibbledmean(ttsum, var_num)
        return var_num + 1

    while stepsize >= 2:
        var_num = fillsquares(var_num)
        var_num = filldiamonds(var_num)
        stepsize //= 2
        wibble /= wibbledecay

    maparray = maparray - maparray.min()
    return (maparray / maparray.max()).reshape(bsize, 1, mapsize, mapsize)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
        name='advex-uar',
        version='0.0.4.dev',
        author='Daniel Kang',
        author_email='daniel.d.kang@gmail.com',
        url='https://github.com/ddkang/advex-uar/',
        packages=setuptools.find_packages(),
        license='Apache',
)



from .elastic_attack import ElasticAttack
from .fw_attack import FrankWolfeAttack
from .jpeg_attack import JPEGAttack
from .pgd_attack import PGDAttack
from .gabor_attack import GaborAttack
from .fog_attack import FogAttack
from .snow_attack import SnowAttack




import torch
import torch.nn as nn
import torchvision

IMAGENET_MEAN = [0.485, 0.456, 0.406]
IMAGENET_STD = [0.229, 0.224, 0.225]

def get_eps_params(base_eps, resol):
    eps_list = []
    max_list = []
    min_list = []
    for i in range(3):
        eps_list.append(torch.full((resol, resol), base_eps, device='cuda'))
        min_list.append(torch.full((resol, resol), 0., device='cuda'))
        max_list.append(torch.full((resol, resol), 255., device='cuda'))

    eps_t = torch.unsqueeze(torch.stack(eps_list), 0)
    max_t = torch.unsqueeze(torch.stack(max_list), 0)
    min_t = torch.unsqueeze(torch.stack(min_list), 0)
    return eps_t, max_t, min_t

def get_imagenet_params(resol):
    mean_list = []
    std_list = []
    for i in range(3):
        mean_list.append(torch.full((resol, resol), IMAGENET_MEAN[i], device='cuda'))
        std_list.append(torch.full((resol, resol), IMAGENET_STD[i], device='cuda'))
    return torch.unsqueeze(torch.stack(mean_list), 0), torch.unsqueeze(torch.stack(std_list), 0)

class ImagenetTransform(nn.Module):
    def __init__(self, resol):
        super().__init__()
        self.mean, self.std = get_imagenet_params(resol)
        
    def forward(self, x):
        '''
        Parameters:
            x: input image with pixels normalized to [0, 255]
        '''
        x = x / 255.
        x = x.sub(self.mean)
        x = x.div(self.std)
        return x

class InverseImagenetTransform(nn.Module):
    def __init__(self, resol):
        super().__init__()
        self.mean, self.std = get_imagenet_params(resol)

    def forward(self, x):
        '''
        Parameters:
            x: input image with pixels normalized to ([0, 1] - IMAGENET_MEAN) / IMAGENET_STD
        '''
        x = x.mul(self.std)
        x = x.add(self.mean)
        x = x * 255.
        return x

class PixelModel(nn.Module):
    def __init__(self, model, resol):
        super().__init__()
        self.model = model
        self.transform = ImagenetTransform(resol)

    def forward(self, x):
        '''
        Parameters:
            x: input image with pixels normalized to [0, 255]
        '''
        x = self.transform(x)
        # x is now normalized as the model expects
        x = self.model(x)
        return x

class AttackWrapper(nn.Module):
    def __init__(self, resol):
        super().__init__()
        self.resol = resol
        self.transform = ImagenetTransform(resol)
        self.inverse_transform = InverseImagenetTransform(resol)
        self.epoch = 0
        
    def forward(self, model, img, *args, **kwargs):
        was_training = model.training
        pixel_model = PixelModel(model, self.resol)
        pixel_model.eval()
        pixel_img = self.inverse_transform(img.detach())
        pixel_ret = self._forward(pixel_model, pixel_img, *args, **kwargs)
        if was_training:
            pixel_model.train()
        ret = self.transform(pixel_ret)
        return ret

    def set_epoch(self, epoch):
        self.epoch = epoch
        self._update_params(epoch)

    def _update_params(self, epoch):
        pass
